# -*- coding: utf-8 -*-
"""
    verktyg.application
    ~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import sys

from werkzeug import Request
from werkzeug.local import Local, LocalManager
from werkzeug.utils import cached_property

from verktyg.exception_dispatch import (
    ExceptionDispatcher, ExceptionHandler
)
from verktyg.routing import URLMap, Route
from verktyg.dispatch import Dispatcher
from verktyg.views import expose


class Application(object):
    """ helper class for creating a wsgi application from a url map and list of
    bindings.

    `url_map`
        werkzeug `URLMap` object that maps from urls to names

    `dispatcher`
        object to map from endpoint names to handler functions

    """

    #: Constructor applied to each wsgi environment to create the request
    #: object to be passed to the handler
    request_class = Request

    @cached_property
    def url_map(self):
        return URLMap()

    @cached_property
    def dispatcher(self):
        return Dispatcher()

    @cached_property
    def exception_dispatcher(self):
        return ExceptionDispatcher()

    def __init__(self, debug=False):
        # Application.__setattr__ depends on _properties so we need to set it
        # using the parent implementation.  A bit magic
        super(Application, self).__setattr__('_properties', {})
        self._methods = {}

        self.debug = debug

        # reference to the bottom of a stack of wsgi middleware wrapping
        # :method:`_dispatch_request`. Invoked by :method:`__call__`.
        # Essentially the real wsgi application.
        self._stack = self._dispatch_request

        # TODO provide a way of adding request specific variables.  Need to be
        # able to register name, `(Application, wsgi_env) -> value` pairs
        # Alternatively get rid of this entirely as it's a massive hack
        self._local = Local()
        self._wsgi_env = self._local('wsgi_env')
        self._map_adapter = self._local('map_adapter')

        local_manager = LocalManager([self._local])
        self.add_middleware(local_manager.make_middleware)

    def add_routes(self, *routes):
        self.url_map.add_routes(*routes)

    def add_bindings(self, *views):
        self.dispatcher.add_bindings(*views)

    def expose(self, endpoint=None, *args, **kwargs):
        """ Decorator to bind a function to an endpoint and optionally the
        endpoint to a route.
        """
        def wrapper(f):
            # nonlocal workaround
            endpoint_ = endpoint
            if endpoint_ is None:
                endpoint_ = f.__name__

            route = kwargs.pop('route', None)
            if route is not None:
                self.add_routes(Route(route, endpoint=endpoint_))

            return expose(self.dispatcher, endpoint_, *args, **kwargs)(f)
        return wrapper

    def add_middleware(self, middleware, *args, **kwargs):
        """ Wrap the application in a layer of wsgi middleware.

        :param middleware:
            a function which takes a wsgi application as it's first argument
            and returns a new wsgi application.  Any other args or kwargs are
            passed after.
        """
        self._stack = middleware(self._stack, *args, **kwargs)

    def add_exception_handler(self, exception_class, handler, **kwargs):
        """ Bind a function to render exceptions of the given class and all
        sub classes.

        Exception handlers take three arguments:
          * a reference to the application
          * a request object
          * the class of the exception
          * the exception instance
          * a traceback
        The last three arguments are the same as the return value of
        `sys.exc_info()`
        """
        self.exception_dispatcher.add_exception_handlers(
            ExceptionHandler(exception_class, handler, **kwargs)
        )

    def exception_handler(self, exception_class, **kwargs):
        """ Decorator that can be used to bind an exception handler to the
        application.  Takes the same arguments as `ExceptionHandler`
        """
        def wrapper(handler):
            self.add_exception_handler(exception_class, handler, **kwargs)
            return handler
        return wrapper

    def __getattr__(self, name):
        if name in self._properties:
            geter, seter = self._properties[name]
            return geter(self)
        elif name in self._methods:
            method = self._methods[name]
            return lambda *args, **kwargs: method(self, *args, **kwargs)
        else:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        if name in self._properties:
            geter, seter = self._properties[name]
            if seter is None:
                raise AttributeError("can't set attribute")
            seter(self, value)
        else:
            super(Application, self).__setattr__(name, value)

    def add_property(self, name, geter, seter=None):
        self._properties[name] = geter, seter

    def add_method(self, name, method):
        """ Bind function as a method of the application

        When the method is called, a `self` parameter will be prepended to it's
        list of arguments.
        """
        self._methods[name] = method

    def _bind(self, wsgi_env):
        self._local.wsgi_env = wsgi_env
        self._local.map_adapter = self.url_map.bind_to_environ(wsgi_env)

    def _dispatch_request(self, wsgi_env, start_response):
        self._bind(wsgi_env)

        request = self.request_class(wsgi_env)

        def call_view(name, kwargs):
            endpoint = self.dispatcher.lookup(
                name,
                method=wsgi_env.get('REQUEST_METHOD'),
                accept=wsgi_env.get('HTTP_ACCEPT'),
                accept_charset=wsgi_env.get('HTTP_ACCEPT_CHARSET'),
                accept_language=wsgi_env.get('HTTP_ACCEPT_LANGUAGE')
            )

            return endpoint(self, request, **kwargs)

        try:
            response = self._map_adapter.dispatch(call_view)
        except BaseException:
            type_, value_, traceback_ = sys.exc_info()

            handler = self.exception_dispatcher.lookup(
                type_,
                accept=wsgi_env.get('HTTP_ACCEPT'),
                accept_charset=wsgi_env.get('HTTP_ACCEPT_CHARSET'),
                accept_language=wsgi_env.get('HTTP_ACCEPT_LANGUAGE')
            )
            if handler is None:
                raise

            response = handler(self, request, type_, value_, traceback_)

        return response(wsgi_env, start_response)

    def __call__(self, wsgi_env, start_response):
        return self._stack(wsgi_env, start_response)

    def url_for(self, endpoint, **kwargs):
        """ construct the url corresponding to an endpoint name and parameters

        Unfortunately will only work if the application has been bound to a
        wsgi request.  If it is not then there is not generally enough
        information to construct full urls.  TODO.
        """
        return self._map_adapter.build(endpoint, values=kwargs)



"""
    verktyg.testsuite
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Contains all Dispatcher tests.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

import werkzeug.testsuite as wzt


class BetterLoader(wzt.BetterLoader):
    """Werkzeug loader saves werkzeug root suite in closure.  This just
    overrides it.
    """
    def getRootSuite(self):
        return suite()


def suite():
    suite = unittest.TestSuite()
    for other_suite in wzt.iter_suites(__name__):
        suite.addTest(other_suite)
    return suite


def main():
    try:
        unittest.main(testLoader=BetterLoader(), defaultTest='suite')
    except Exception:
        import sys
        import traceback
        traceback.print_exc()
        sys.exit(1)



# -*- coding: utf-8 -*-
"""
    verktyg.dispatch
    ~~~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
from werkzeug.exceptions import NotImplemented, MethodNotAllowed

from verktyg.accept import Representation, select_representation


class BindingFactory(object):
    def get_bindings(self):
        raise NotImplementedError()


class Binding(BindingFactory, Representation):
    """Represents an action associated with a single combination of endpoint
    name, method and content-type.  In most cases you probably want to be using
    subclasses of `View` which can listen for multiple methods.

    `name`
        A hashable identifier.

    `method`
        An http method as an upper case string.

    `action`
        The action to perform if the binding is matched

    `content_type`
        A type/subtype formatted string representing the content type that the
        handler returns
        See `Representation`

    `language`
        See `Representation`

    `charser`
        See `Representation`

    `qs`
        Quality of source.  Multiplied by the accept q value to give quality of
        binding if mimetypes match.  Name by convention from other servers
    """
    def __init__(self, name, action, method='GET', **kwargs):
        self.name = name
        self.method = method
        self.action = action

        super(Binding, self).__init__(**kwargs)

    def get_bindings(self):
        yield self

    def __repr__(self):
        return '<%s %s %s %s>' % (
            self.__class__.__name__,
            repr(self.name),
            repr(self.method),
            repr(self.content_type),
        )


class Dispatcher(BindingFactory):
    def __init__(self, views=[]):
        self._index = {}
        self._views = []

        for view in views:
            self.add_bindings(view)

    def add_bindings(self, *factories):
        """ Add bindings from bindings factory to this dispatcher.
        Dispatchers can be nested
        """
        for factory in factories:
            for view in factory.get_bindings():
                if view.name not in self._index:
                    self._index[view.name] = {}
                with_name = self._index[view.name]

                if view.method not in with_name:
                    with_name[view.method] = []
                with_method = with_name[view.method]

                with_method.append(view)

                self._views.append(view)

    def get_bindings(self):
        return iter(self._views)

    def lookup(self, name, method='GET',
               accept='*/*', accept_language=None, accept_charset=None):
        with_name = self._index.get(name)
        if name not in self._index:
            raise NotImplemented()

        with_method = with_name.get(method)
        if with_method is None:
            if method == 'HEAD' and 'GET' in with_name:
                with_method = with_name['GET']
            else:
                raise MethodNotAllowed()

        representation = select_representation(
            with_method,
            accept=accept,
            accept_language=accept_language,
            accept_charset=accept_charset
        )

        return representation.action



import jinja2

from werkzeug.utils import cached_property


class Jinja2ApplicationMixin(object):
    @cached_property
    def jinja_env(self):
        return jinja2.Environment(loader=jinja2.ChoiceLoader([]))

    def add_templates(self, loader):
        self.jinja_env.loader.loaders.append(loader)

    def get_renderer(self, name):
        if isinstance(name, str):
            return self.jinja_env.get_template(name).render
        else:
            # assume callable
            return name



# -*- coding: utf-8 -*-
"""
    verktyg
    ~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
from verktyg.routing import URLMap
from verktyg.dispatch import Dispatcher
from verktyg.application import Application

__all__ = ['URLMap', 'Dispatcher', 'Application']



# -*- coding: utf-8 -*-
"""
    verktyg.views
    ~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import json

from werkzeug import Response
from verktyg.dispatch import BindingFactory, Binding


class View(BindingFactory):
    """ Wraps a function or callable so that it can be bound to a name in a
    dispatcher.
    """
    def __init__(self, name, action, methods=None, content_type=None, qs=None):
        self._name = name

        if methods is None:
            self._methods = set(['GET'])
        elif isinstance(methods, str):
            self._methods = set([methods])
        else:
            self._methods = methods

        self._content_type = content_type
        self._qs = qs
        self._action = action

    def __call__(self, env, req, *args, **kwargs):
        return self._action(env, req, *args, **kwargs)

    def get_bindings(self):
        for method in self._methods:
            yield Binding(self._name, self,
                          method=method,
                          content_type=self._content_type)


class ClassView(BindingFactory):
    def get_bindings(self):
        for method in set(['GET', 'HEAD', 'POST', 'PUT', 'DELETE']):  # TODO
            if hasattr(self, method):
                yield Binding(self.name, getattr(self, method),
                              method=method)


class TemplateView(View):
    """ Like `View` but if the value returned from the action is not an
    instance of `Response` it is rendered using the named template.

    :param name:
    :param action: called with environment, request and params to generate
                   response.  See `View`.
    :param template: either a string naming the template to be retrieved from
                     the environment or a callable applied to the result to
                     create an http `Response` object
    """
    def __init__(self, name, action,
                 methods=None, template=None,
                 content_type=None):
        super(TemplateView, self).__init__(
            name, action,
            methods=methods,
            content_type=content_type)
        self._template = template

    def __call__(self, env, req, *args, **kwargs):
        res = self._action(env, req, *args, **kwargs)
        if isinstance(res, Response):
            return res
        return Response(env.get_renderer(self._template)(res))


class JsonView(View):
    def __init__(self, name, action, methods=None, qs=None):
        super(JsonView, self).__init__(
            name, action, methods=methods,
            content_type='application/json', qs=qs
        )

    def __call__(self, env, req, *args, **kwargs):
        res = super(JsonView, self).__call__(env, req, *args, **kwargs)

        if isinstance(res, Response):
            # rendering already done
            return res

        if res is None:
            # no content
            return Response(status=204)

        return Response(json.dumps(res), content_type='application/json')


def expose(dispatcher, name, *args, **kwargs):
    def decorator(f):
        dispatcher.add_bindings(TemplateView(name, f, *args, **kwargs))
        return f
    return decorator


def expose_html(*args, **kwargs):
    if 'content_type' not in kwargs:
        kwargs['content_type'] = 'text/html'
    return expose(*args, **kwargs)


def expose_json(dispatcher, name, *args, **kwargs):
    def decorator(f):
        dispatcher.add_bindings(JsonView(name, f, *args, **kwargs))
        return f
    return decorator



"""
    verktyg.testsuite.dispatch
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for the core dispatcher.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from werkzeug.testsuite import WerkzeugTestCase

from werkzeug.exceptions import NotImplemented, MethodNotAllowed, NotAcceptable

from verktyg.dispatch import Binding, Dispatcher


class DispatchTestCase(WerkzeugTestCase):
    def test_name_dispatch(self):
        dispatcher = Dispatcher([
            Binding('tweedle-dum', 'Tweedle Dum'),
            Binding('tweedle-dee', 'Tweedle Dee'),
            Binding('same', 'overridden'),
            Binding('same', 'overriding'),
            ])

        self.assert_equal('Tweedle Dum', dispatcher.lookup('tweedle-dum'))
        self.assert_equal('Tweedle Dee', dispatcher.lookup('tweedle-dee'))
        self.assert_raises(NotImplemented, dispatcher.lookup, 'non-existant')
        self.assert_equal('overriding', dispatcher.lookup('same'))

    def test_method_dispatch(self):
        dispatcher = Dispatcher([
            Binding('test', 'get', method='GET'),
            Binding('test', 'post', method='POST'),
            Binding('head', 'head', method='HEAD'),
            Binding('no-head', 'get', method='GET'),
            ])

        # default to 'GET'
        self.assert_equal('get',
                          dispatcher.lookup('test'))
        self.assert_equal('get',
                          dispatcher.lookup('test', method='GET'))

        # `POST` gives something different
        self.assert_equal('post',
                          dispatcher.lookup('test', method='POST'))

        # `PUT` not found
        self.assert_raises(MethodNotAllowed,
                           dispatcher.lookup, 'test', method='PUT')

    def test_head_fallback(self):
        dispatcher = Dispatcher([
            Binding('head', 'head', method='HEAD'),
            Binding('no-head', 'get', method='GET'),
            ])

        # `HEAD` should fall back to `GET`
        self.assert_equal('head',
                          dispatcher.lookup('head', method='HEAD'))
        self.assert_equal('get',
                          dispatcher.lookup('no-head', method='HEAD'))

    def test_method_override(self):
        dispatcher = Dispatcher([
            Binding('same', 'overridden'),
            Binding('same', 'unaffected', method='POST'),
            Binding('same', 'overriding'),
            ])

        # replacing handler for one method should not affect others
        self.assert_equal('overriding',
                          dispatcher.lookup('same'))
        self.assert_equal('unaffected',
                          dispatcher.lookup('same', method='POST'))

    def test_accept_dispatch(self):
        dispatcher = Dispatcher([
            Binding('test', 'json', content_type='application/json'),
            Binding('test', 'html', content_type='text/html'),
            Binding('test', 'whatever'),
            Binding('nope', 'nope', content_type='application/xml'),
            ])

        # accept header strings
        self.assert_equal(
            'json',
            dispatcher.lookup('test', accept='application/json')
        )

        self.assert_equal(
            'json',
            dispatcher.lookup(
                'test', accept='application/json; q=0.9, text/html; q=0.8'
            )
        )
        self.assert_equal(
            'whatever',
            dispatcher.lookup('test', accept='application/xml')
        )

        self.assert_raises(
            NotAcceptable,
            dispatcher.lookup, 'nope', accept='text/html')

    def test_nested(self):
        child = Dispatcher([
            Binding('nested', 'Nested'),
            ])
        parent = Dispatcher([
            child,
            ])

        self.assert_equal(
            'Nested',
            parent.lookup('nested'))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(DispatchTestCase))
    return suite



# -*- coding: utf-8 -*-
"""
    verktyg.accept
    ~~~~~~~~~~~~~~

    Code for selecting handlers based on accept headers

    http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import mimeparse

from werkzeug.exceptions import NotAcceptable


def select_representation(
        representations,
        accept='*/*', accept_language=None, accept_charset=None):
    max_quality = tuple()
    best = None
    for representation in representations:
        try:
            quality = representation.quality(accept=accept,
                                             accept_language=accept_language,
                                             accept_charset=accept_charset)
        except NotAcceptable:
            continue

        if not isinstance(quality, tuple):
            quality = (quality,)

        # Later bindings take precedence
        if quality >= max_quality:
            best = representation
            max_quality = quality

    if best is None:
        raise NotAcceptable()

    return best


class Representation(object):
    def __init__(self, content_type=None, language=None,
                 charset=None, qs=None):
        self.content_type = content_type
        self.language = language
        self.charset = charset

        if qs is None:
            if content_type is None:
                self.qs = 0.001
            else:
                self.qs = 1.0

    def quality(self, accept=None, accept_charset=None,
                accept_language=None):
        """
        :param accept: string in the same format as an http `Accept` header

        :param accept_language: string in the same format as an http
            `Accept-Language` header

        :param accept_charset: string in the same format as an http
            `Accept-Charset` header

        :return: a number or tuple of tuples representing the quality of
            the match. By convention outer tuples should be in content type,
            language, charset order.  Raises `NotAcceptable If the binding does
            not match the request.

        """
        if self.content_type is None:
            # TODO
            return 5, self.qs

        if accept is None:
            return 0, self.qs

        accept = [
            mimeparse.parse_media_range(media_range)
            for media_range in accept.split(',')
        ]

        fitness, quality = mimeparse.fitness_and_quality_parsed(
            self.content_type, accept
        )

        if fitness == -1:
            raise NotAcceptable()

        return fitness, quality * self.qs



# -*- coding: utf-8 -*-
"""
    verktyg.testsuite.routing
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Routing tests.

    :copyright:
        (c) 2014 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import unittest
import uuid

from werkzeug.testsuite import WerkzeugTestCase

from verktyg import routing as r
from werkzeug.wrappers import Response
from werkzeug.datastructures import ImmutableDict
from werkzeug.test import create_environ


class RoutingTestCase(WerkzeugTestCase):

    def test_basic_routing(self):
        map = r.URLMap([
            r.Route('/', endpoint='index'),
            r.Route('/foo', endpoint='foo'),
            r.Route('/bar/', endpoint='bar')
        ])
        adapter = map.bind('example.org', '/')
        self.assert_equal(
            adapter.match('/'),
            ('index', {})
        )
        self.assert_equal(
            adapter.match('/foo'),
            ('foo', {})
        )
        self.assert_equal(
            adapter.match('/bar/'),
            ('bar', {})
        )
        self.assert_raises(r.RequestRedirect, lambda: adapter.match('/bar'))
        self.assert_raises(r.NotFound, lambda: adapter.match('/blub'))

        adapter = map.bind('example.org', '/test')
        try:
            adapter.match('/bar')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://example.org/test/bar/')
        else:
            self.fail('Expected request redirect')

        adapter = map.bind('example.org', '/')
        try:
            adapter.match('/bar')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://example.org/bar/')
        else:
            self.fail('Expected request redirect')

        adapter = map.bind('example.org', '/')
        try:
            adapter.match('/bar', query_args={'aha': 'muhaha'})
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://example.org/bar/?aha=muhaha')
        else:
            self.fail('Expected request redirect')

        adapter = map.bind('example.org', '/')
        try:
            adapter.match('/bar', query_args='aha=muhaha')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://example.org/bar/?aha=muhaha')
        else:
            self.fail('Expected request redirect')

        adapter = map.bind_to_environ(create_environ('/bar?foo=bar',
                                                     'http://example.org/'))
        try:
            adapter.match()
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://example.org/bar/?foo=bar')
        else:
            self.fail('Expected request redirect')

    def test_environ_defaults(self):
        environ = create_environ("/foo")
        self.assert_strict_equal(environ["PATH_INFO"], '/foo')
        m = r.URLMap([
            r.Route("/foo", endpoint="foo"),
            r.Route("/bar", endpoint="bar"),
        ])
        a = m.bind_to_environ(environ)
        self.assert_strict_equal(a.match("/foo"), ('foo', {}))
        self.assert_strict_equal(a.match(), ('foo', {}))
        self.assert_strict_equal(a.match("/bar"), ('bar', {}))
        self.assert_raises(r.NotFound, a.match, "/bars")

    def test_environ_nonascii_pathinfo(self):
        environ = create_environ(u'/лошадь')
        m = r.URLMap([
            r.Route(u'/', endpoint='index'),
            r.Route(u'/лошадь', endpoint='horse')
        ])
        a = m.bind_to_environ(environ)
        self.assert_strict_equal(a.match(u'/'), ('index', {}))
        self.assert_strict_equal(a.match(u'/лошадь'), ('horse', {}))
        self.assert_raises(r.NotFound, a.match, u'/барсук')

    def test_basic_building(self):
        map = r.URLMap([
            r.Route('/', endpoint='index'),
            r.Route('/foo', endpoint='foo'),
            r.Route('/bar/<baz>', endpoint='bar'),
            r.Route('/bar/<int:bazi>', endpoint='bari'),
            r.Route('/bar/<float:bazf>', endpoint='barf'),
            r.Route('/bar/<path:bazp>', endpoint='barp'),
            r.Route('/hehe', endpoint='blah', subdomain='blah')
        ])
        adapter = map.bind('example.org', '/', subdomain='blah')

        self.assert_equal(
            adapter.build('index', {}),
            'http://example.org/'
        )
        self.assert_equal(
            adapter.build('foo', {}),
            'http://example.org/foo'
        )
        self.assert_equal(
            adapter.build('bar', {'baz': 'blub'}),
            'http://example.org/bar/blub'
        )
        self.assert_equal(
            adapter.build('bari', {'bazi': 50}),
            'http://example.org/bar/50'
        )
        self.assert_equal(
            adapter.build('barf', {'bazf': 0.815}),
            'http://example.org/bar/0.815'
        )
        self.assert_equal(
            adapter.build('barp', {'bazp': 'la/di'}),
            'http://example.org/bar/la/di'
        )
        self.assert_equal(
            adapter.build('blah', {}),
            '/hehe'
        )
        self.assert_raises(r.BuildError, lambda: adapter.build('urks'))

        adapter = map.bind('example.org', '/test', subdomain='blah')
        self.assert_equal(
            adapter.build('index', {}),
            'http://example.org/test/'
        )
        self.assert_equal(
            adapter.build('foo', {}),
            'http://example.org/test/foo'
        )
        self.assert_equal(
            adapter.build('bar', {'baz': 'blub'}),
            'http://example.org/test/bar/blub'
        )
        self.assert_equal(
            adapter.build('bari', {'bazi': 50}),
            'http://example.org/test/bar/50'
        )
        self.assert_equal(
            adapter.build('barf', {'bazf': 0.815}),
            'http://example.org/test/bar/0.815'
        )
        self.assert_equal(
            adapter.build('barp', {'bazp': 'la/di'}),
            'http://example.org/test/bar/la/di'
        )
        self.assert_equal(
            adapter.build('blah', {}),
            '/test/hehe'
        )

    def test_defaults(self):
        map = r.URLMap([
            r.Route('/foo/', defaults={'page': 1}, endpoint='foo'),
            r.Route('/foo/<int:page>', endpoint='foo')
        ])
        adapter = map.bind('example.org', '/')

        self.assert_equal(
            adapter.match('/foo/'),
            ('foo', {'page': 1})
        )
        self.assert_raises(r.RequestRedirect, lambda: adapter.match('/foo/1'))
        self.assert_equal(
            adapter.match('/foo/2'),
            ('foo', {'page': 2})
        )
        self.assert_equal(adapter.build('foo', {}), '/foo/')
        self.assert_equal(adapter.build('foo', {'page': 1}), '/foo/')
        self.assert_equal(adapter.build('foo', {'page': 2}), '/foo/2')

    def test_greedy(self):
        map = r.URLMap([
            r.Route('/foo', endpoint='foo'),
            r.Route('/<path:bar>', endpoint='bar'),
            r.Route('/<path:bar>/<path:blub>', endpoint='bar')
        ])
        adapter = map.bind('example.org', '/')

        self.assert_equal(
            adapter.match('/foo'),
            ('foo', {})
        )
        self.assert_equal(
            adapter.match('/blub'),
            ('bar', {'bar': 'blub'})
        )
        self.assert_equal(
            adapter.match('/he/he'),
            ('bar', {'bar': 'he', 'blub': 'he'})
        )

        self.assert_equal(adapter.build('foo', {}), '/foo')
        self.assert_equal(adapter.build('bar', {'bar': 'blub'}), '/blub')
        self.assert_equal(
            adapter.build('bar', {'bar': 'blub', 'blub': 'bar'}),
            '/blub/bar'
        )

    def test_path(self):
        map = r.URLMap([
            r.Route('/', defaults={'name': 'FrontPage'}, endpoint='page'),
            r.Route('/Special', endpoint='special'),
            r.Route('/<int:year>', endpoint='year'),
            r.Route('/<path:name>', endpoint='page'),
            r.Route('/<path:name>/edit', endpoint='editpage'),
            r.Route(
                '/<path:name>/silly/<path:name2>',
                endpoint='sillypage'
            ),
            r.Route(
                '/<path:name>/silly/<path:name2>/edit',
                endpoint='editsillypage'
            ),
            r.Route('/Talk:<path:name>', endpoint='talk'),
            r.Route('/User:<username>', endpoint='user'),
            r.Route('/User:<username>/<path:name>', endpoint='userpage'),
            r.Route('/Files/<path:file>', endpoint='files'),
        ])
        adapter = map.bind('example.org', '/')

        self.assert_equal(
            adapter.match('/'),
            ('page', {'name': 'FrontPage'})
        )
        self.assert_raises(
            r.RequestRedirect,
            lambda: adapter.match('/FrontPage')
        )
        self.assert_equal(
            adapter.match('/Special'),
            ('special', {})
        )
        self.assert_equal(
            adapter.match('/2007'),
            ('year', {'year': 2007})
        )
        self.assert_equal(
            adapter.match('/Some/Page'),
            ('page', {'name': 'Some/Page'})
        )
        self.assert_equal(
            adapter.match('/Some/Page/edit'),
            ('editpage', {'name': 'Some/Page'})
        )
        self.assert_equal(
            adapter.match('/Foo/silly/bar'),
            ('sillypage', {'name': 'Foo', 'name2': 'bar'})
        )
        self.assert_equal(
            adapter.match('/Foo/silly/bar/edit'),
            ('editsillypage', {'name': 'Foo', 'name2': 'bar'})
        )
        self.assert_equal(
            adapter.match('/Talk:Foo/Bar'),
            ('talk', {'name': 'Foo/Bar'})
        )
        self.assert_equal(
            adapter.match('/User:thomas'),
            ('user', {'username': 'thomas'})
        )
        self.assert_equal(
            adapter.match('/User:thomas/projects/werkzeug'),
            ('userpage', {'username': 'thomas', 'name': 'projects/werkzeug'})
        )
        self.assert_equal(
            adapter.match('/Files/downloads/werkzeug/0.2.zip'),
            ('files', {'file': 'downloads/werkzeug/0.2.zip'})
        )

    def test_dispatch(self):
        env = create_environ('/')
        map = r.URLMap([
            r.Route('/', endpoint='root'),
            r.Route('/foo/', endpoint='foo')
        ])
        adapter = map.bind_to_environ(env)

        raise_this = None

        def view_func(endpoint, values):
            if raise_this is not None:
                raise raise_this
            return Response(repr((endpoint, values)))

        def dispatch(p, q=False):
            return Response.force_type(
                adapter.dispatch(view_func, p, catch_http_exceptions=q),
                env
            )

        self.assert_equal(dispatch('/').data, b"('root', {})")
        self.assert_equal(dispatch('/foo').status_code, 301)
        raise_this = r.NotFound()
        self.assert_raises(r.NotFound, lambda: dispatch('/bar'))
        self.assert_equal(dispatch('/bar', True).status_code, 404)

    def test_http_host_before_server_name(self):
        env = {
            'HTTP_HOST':            'wiki.example.com',
            'SERVER_NAME':          'web0.example.com',
            'SERVER_PORT':          '80',
            'SCRIPT_NAME':          '',
            'PATH_INFO':            '',
            'REQUEST_METHOD':       'GET',
            'wsgi.url_scheme':      'http'
        }
        map = r.URLMap([r.Route('/', endpoint='index', subdomain='wiki')])
        adapter = map.bind_to_environ(env, server_name='example.com')
        self.assert_equal(
            adapter.match('/'),
            ('index', {})
        )
        self.assert_equal(
            adapter.build('index', force_external=True),
            'http://wiki.example.com/'
        )
        self.assert_equal(
            adapter.build('index'),
            '/'
        )

        env['HTTP_HOST'] = 'admin.example.com'
        adapter = map.bind_to_environ(env, server_name='example.com')
        self.assert_equal(adapter.build('index'), 'http://wiki.example.com/')

    def test_adapter_url_parameter_sorting(self):
        map = r.URLMap(
            [r.Route('/', endpoint='index')],
            sort_parameters=True,
            sort_key=lambda x: x[1]
        )
        adapter = map.bind('localhost', '/')
        self.assert_equal(
            adapter.build('index', {'x': 20, 'y': 10, 'z': 30},
                          force_external=True),
            'http://localhost/?y=10&x=20&z=30'
        )

    def test_request_direct_charset_bug(self):
        map = r.URLMap([r.Route(u'/öäü/')])
        adapter = map.bind('localhost', '/')
        try:
            adapter.match(u'/öäü')
        except r.RequestRedirect as e:
            self.assert_equal(
                e.new_url,
                'http://localhost/%C3%B6%C3%A4%C3%BC/'
            )
        else:
            self.fail('expected request redirect exception')

    def test_request_redirect_default(self):
        map = r.URLMap([
            r.Route(u'/foo', defaults={'bar': 42}),
            r.Route(u'/foo/<int:bar>'),
        ])
        adapter = map.bind('localhost', '/')
        try:
            adapter.match(u'/foo/42')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://localhost/foo')
        else:
            self.fail('expected request redirect exception')

    def test_request_redirect_default_subdomain(self):
        map = r.URLMap([
            r.Route(u'/foo', defaults={'bar': 42}, subdomain='test'),
            r.Route(u'/foo/<int:bar>', subdomain='other'),
        ])
        adapter = map.bind('localhost', '/', subdomain='other')
        try:
            adapter.match(u'/foo/42')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://test.localhost/foo')
        else:
            self.fail('expected request redirect exception')

    def test_adapter_match_return_route(self):
        route = r.Route('/foo/', endpoint='foo')
        map = r.URLMap([route])
        adapter = map.bind('localhost', '/')
        self.assert_equal(
            adapter.match('/foo/', return_route=True),
            (route, {})
        )

    def test_server_name_interpolation(self):
        server_name = 'example.invalid'
        map = r.URLMap([
            r.Route('/', endpoint='index'),
            r.Route('/', endpoint='alt', subdomain='alt'),
        ])

        env = create_environ('/', 'http://%s/' % server_name)
        adapter = map.bind_to_environ(env, server_name=server_name)
        self.assert_equal(
            adapter.match(),
            ('index', {})
        )

        env = create_environ('/', 'http://alt.%s/' % server_name)
        adapter = map.bind_to_environ(env, server_name=server_name)
        self.assert_equal(
            adapter.match(),
            ('alt', {})
        )

        env = create_environ('/', 'http://%s/' % server_name)
        adapter = map.bind_to_environ(env, server_name='foo')
        self.assert_equal(adapter.subdomain, '<invalid>')

    def test_route_templates(self):
        testcase = r.RouteTemplate([
            r.Submount('/test/$app', [
                r.Route('/foo/', endpoint='handle_foo'),
                r.Route('/bar/', endpoint='handle_bar'),
                r.Route('/baz/', endpoint='handle_baz'),
            ]),
            r.EndpointPrefix('${app}', [
                r.Route('/${app}-blah', endpoint='bar'),
                r.Route('/${app}-meh', endpoint='baz'),
            ]),
            r.Subdomain('$app', [
                r.Route('/blah', endpoint='x_bar'),
                r.Route('/meh', endpoint='x_baz'),
            ])
        ])

        url_map = r.URLMap([
            testcase(app='test1'),
            testcase(app='test2'),
            testcase(app='test3'),
            testcase(app='test4'),
        ])

        out = sorted([(x.route, x.subdomain, x.endpoint)
                      for x in url_map.iter_routes()])

        self.assert_equal(out, [
            ('/blah', 'test1', 'x_bar'),
            ('/blah', 'test2', 'x_bar'),
            ('/blah', 'test3', 'x_bar'),
            ('/blah', 'test4', 'x_bar'),
            ('/meh', 'test1', 'x_baz'),
            ('/meh', 'test2', 'x_baz'),
            ('/meh', 'test3', 'x_baz'),
            ('/meh', 'test4', 'x_baz'),
            ('/test/test1/bar/', '', 'handle_bar'),
            ('/test/test1/baz/', '', 'handle_baz'),
            ('/test/test1/foo/', '', 'handle_foo'),
            ('/test/test2/bar/', '', 'handle_bar'),
            ('/test/test2/baz/', '', 'handle_baz'),
            ('/test/test2/foo/', '', 'handle_foo'),
            ('/test/test3/bar/', '', 'handle_bar'),
            ('/test/test3/baz/', '', 'handle_baz'),
            ('/test/test3/foo/', '', 'handle_foo'),
            ('/test/test4/bar/', '', 'handle_bar'),
            ('/test/test4/baz/', '', 'handle_baz'),
            ('/test/test4/foo/', '', 'handle_foo'),
            ('/test1-blah', '', 'test1bar'),
            ('/test1-meh', '', 'test1baz'),
            ('/test2-blah', '', 'test2bar'),
            ('/test2-meh', '', 'test2baz'),
            ('/test3-blah', '', 'test3bar'),
            ('/test3-meh', '', 'test3baz'),
            ('/test4-blah', '', 'test4bar'),
            ('/test4-meh', '', 'test4baz')
        ])

    def test_non_string_parts(self):
        m = r.URLMap([
            r.Route('/<foo>', endpoint='foo')
        ])
        a = m.bind('example.com')
        self.assert_equal(a.build('foo', {'foo': 42}), '/42')

    def test_complex_routing_routes(self):
        m = r.URLMap([
            r.Route('/', endpoint='index'),
            r.Route('/<int:blub>', endpoint='an_int'),
            r.Route('/<blub>', endpoint='a_string'),
            r.Route('/foo/', endpoint='nested'),
            r.Route('/foobar/', endpoint='nestedbar'),
            r.Route('/foo/<path:testing>/', endpoint='nested_show'),
            r.Route('/foo/<path:testing>/edit', endpoint='nested_edit'),
            r.Route('/users/', endpoint='users', defaults={'page': 1}),
            r.Route('/users/page/<int:page>', endpoint='users'),
            r.Route('/foox', endpoint='foox'),
            r.Route('/<path:bar>/<path:blub>', endpoint='barx_path_path')
        ])
        a = m.bind('example.com')

        self.assert_equal(
            a.match('/'),
            ('index', {})
        )
        self.assert_equal(
            a.match('/42'),
            ('an_int', {'blub': 42})
        )
        self.assert_equal(
            a.match('/blub'),
            ('a_string', {'blub': 'blub'})
        )
        self.assert_equal(
            a.match('/foo/'),
            ('nested', {})
        )
        self.assert_equal(
            a.match('/foobar/'),
            ('nestedbar', {})
        )
        self.assert_equal(
            a.match('/foo/1/2/3/'),
            ('nested_show', {'testing': '1/2/3'})
        )
        self.assert_equal(
            a.match('/foo/1/2/3/edit'),
            ('nested_edit', {'testing': '1/2/3'})
        )
        self.assert_equal(
            a.match('/users/'),
            ('users', {'page': 1})
        )
        self.assert_equal(
            a.match('/users/page/2'),
            ('users', {'page': 2})
        )
        self.assert_equal(
            a.match('/foox'),
            ('foox', {})
        )
        self.assert_equal(
            a.match('/1/2/3'),
            ('barx_path_path', {'bar': '1', 'blub': '2/3'})
        )

        self.assert_equal(
            a.build('index'),
            '/'
        )
        self.assert_equal(
            a.build('an_int', {'blub': 42}),
            '/42'
        )
        self.assert_equal(
            a.build('a_string', {'blub': 'test'}),
            '/test'
        )
        self.assert_equal(
            a.build('nested'),
            '/foo/'
        )
        self.assert_equal(
            a.build('nestedbar'),
            '/foobar/'
        )
        self.assert_equal(
            a.build('nested_show', {'testing': '1/2/3'}),
            '/foo/1/2/3/'
        )
        self.assert_equal(
            a.build('nested_edit', {'testing': '1/2/3'}),
            '/foo/1/2/3/edit'
        )
        self.assert_equal(
            a.build('users', {'page': 1}),
            '/users/'
        )
        self.assert_equal(
            a.build('users', {'page': 2}),
            '/users/page/2'
        )
        self.assert_equal(
            a.build('foox'),
            '/foox'
        )
        self.assert_equal(
            a.build('barx_path_path', {'bar': '1', 'blub': '2/3'}),
            '/1/2/3'
        )

    def test_default_converters(self):
        class MyURLMap(r.URLMap):
            default_converters = r.URLMap.default_converters.copy()
            default_converters['foo'] = r.UnicodeConverter
        assert isinstance(r.URLMap.default_converters, ImmutableDict)
        m = MyURLMap([
            r.Route('/a/<foo:a>', endpoint='a'),
            r.Route('/b/<foo:b>', endpoint='b'),
            r.Route('/c/<c>', endpoint='c')
        ], converters={'bar': r.UnicodeConverter})
        a = m.bind('example.org', '/')
        self.assert_equal(a.match('/a/1'), ('a', {'a': '1'}))
        self.assert_equal(a.match('/b/2'), ('b', {'b': '2'}))
        self.assert_equal(a.match('/c/3'), ('c', {'c': '3'}))
        assert 'foo' not in r.URLMap.default_converters

    def test_uuid_converter(self):
        m = r.URLMap([
            r.Route('/a/<uuid:a_uuid>', endpoint='a')
        ])
        a = m.bind('example.org', '/')
        rooute, kwargs = a.match('/a/a8098c1a-f86e-11da-bd1a-00112444be1e')
        self.assert_equal(type(kwargs['a_uuid']), uuid.UUID)

    def test_build_append_unknown(self):
        map = r.URLMap([
            r.Route('/bar/<float:bazf>', endpoint='barf')
        ])
        adapter = map.bind('example.org', '/', subdomain='blah')
        self.assert_equal(
            adapter.build('barf', {'bazf': 0.815, 'bif': 1.0}),
            'http://example.org/bar/0.815?bif=1.0'
        )
        self.assert_equal(
            adapter.build('barf', {'bazf': 0.815, 'bif': 1.0},
                          append_unknown=False),
            'http://example.org/bar/0.815'
        )

    def test_protocol_joining_bug(self):
        m = r.URLMap([r.Route('/<foo>', endpoint='x')])
        a = m.bind('example.org')
        self.assert_equal(a.build('x', {'foo': 'x:y'}), '/x:y')
        self.assert_equal(
            a.build('x', {'foo': 'x:y'}, force_external=True),
            'http://example.org/x:y'
        )

    def test_external_building_with_port(self):
        map = r.URLMap([
            r.Route('/', endpoint='index'),
        ])
        adapter = map.bind('example.org:5000', '/')
        built_url = adapter.build('index', {}, force_external=True)
        self.assert_equal(built_url, 'http://example.org:5000/', built_url)

    def test_external_building_with_port_bind_to_environ(self):
        map = r.URLMap([
            r.Route('/', endpoint='index'),
        ])
        adapter = map.bind_to_environ(
            create_environ('/', 'http://example.org:5000/'),
            server_name="example.org:5000"
        )
        built_url = adapter.build('index', {}, force_external=True)
        self.assert_equal(built_url, 'http://example.org:5000/', built_url)

    def test_external_building_with_port_bind_to_environ_bad_servername(self):
        map = r.URLMap([
            r.Route('/', endpoint='index'),
        ])
        environ = create_environ('/', 'http://example.org:5000/')
        adapter = map.bind_to_environ(environ, server_name="example.org")
        self.assert_equal(adapter.subdomain, '<invalid>')

    def test_converter_parser(self):
        args, kwargs = r.parse_converter_args(u'test, a=1, b=3.0')

        self.assert_equal(args, ('test',))
        self.assert_equal(kwargs, {'a': 1, 'b': 3.0})

        args, kwargs = r.parse_converter_args('')
        assert not args and not kwargs

        args, kwargs = r.parse_converter_args('a, b, c,')
        self.assert_equal(args, ('a', 'b', 'c'))
        assert not kwargs

        args, kwargs = r.parse_converter_args('True, False, None')
        self.assert_equal(args, (True, False, None))

        args, kwargs = r.parse_converter_args('"foo", u"bar"')
        self.assert_equal(args, ('foo', 'bar'))

    def test_alias_redirects(self):
        m = r.URLMap([
            r.Route('/', endpoint='index'),
            r.Route('/index.html', endpoint='index', alias=True),
            r.Route('/users/', defaults={'page': 1}, endpoint='users'),
            r.Route('/users/index.html', defaults={'page': 1}, alias=True,
                    endpoint='users'),
            r.Route('/users/page/<int:page>', endpoint='users'),
            r.Route('/users/page-<int:page>.html', alias=True,
                    endpoint='users'),
        ])
        a = m.bind('example.com')

        def ensure_redirect(path, new_url, args=None):
            try:
                a.match(path, query_args=args)
            except r.RequestRedirect as e:
                self.assert_equal(e.new_url, 'http://example.com' + new_url)
            else:
                assert False, 'expected redirect'

        ensure_redirect('/index.html', '/')
        ensure_redirect('/users/index.html', '/users/')
        ensure_redirect('/users/page-2.html', '/users/page/2')
        ensure_redirect('/users/page-1.html', '/users/')
        ensure_redirect('/users/page-1.html', '/users/?foo=bar',
                        {'foo': 'bar'})

        self.assert_equal(a.build('index'), '/')
        self.assert_equal(a.build('users', {'page': 1}), '/users/')
        self.assert_equal(a.build('users', {'page': 2}), '/users/page/2')

    def test_double_defaults(self):
        for prefix in '', '/aaa':
            m = r.URLMap([
                r.Route(
                    prefix + '/',
                    defaults={'foo': 1, 'bar': False},
                    endpoint='x'
                ),
                r.Route(
                    prefix + '/<int:foo>',
                    defaults={'bar': False},
                    endpoint='x'
                ),
                r.Route(
                    prefix + '/bar/',
                    defaults={'foo': 1, 'bar': True},
                    endpoint='x'
                ),
                r.Route(
                    prefix + '/bar/<int:foo>',
                    defaults={'bar': True},
                    endpoint='x'
                ),
            ])
            a = m.bind('example.com')

            self.assert_equal(
                a.match(prefix + '/'),
                ('x', {'foo': 1, 'bar': False})
            )
            self.assert_equal(
                a.match(prefix + '/2'),
                ('x', {'foo': 2, 'bar': False})
            )
            self.assert_equal(
                a.match(prefix + '/bar/'),
                ('x', {'foo': 1, 'bar': True})
            )
            self.assert_equal(
                a.match(prefix + '/bar/2'),
                ('x', {'foo': 2, 'bar': True})
            )

            self.assert_equal(
                a.build('x', {'foo': 1, 'bar': False}),
                prefix + '/'
            )
            self.assert_equal(
                a.build('x', {'foo': 2, 'bar': False}),
                prefix + '/2'
            )
            self.assert_equal(
                a.build('x', {'bar': False}),
                prefix + '/'
            )
            self.assert_equal(
                a.build('x', {'foo': 1, 'bar': True}),
                prefix + '/bar/'
            )
            self.assert_equal(
                a.build('x', {'foo': 2, 'bar': True}),
                prefix + '/bar/2'
            )
            self.assert_equal(
                a.build('x', {'bar': True}),
                prefix + '/bar/'
            )

    def test_host_matching(self):
        m = r.URLMap([
            r.Route(
                '/',
                endpoint='index',
                host='www.<domain>'
            ),
            r.Route(
                '/',
                endpoint='files',
                host='files.<domain>'
            ),
            r.Route(
                '/foo/',
                defaults={'page': 1},
                host='www.<domain>',
                endpoint='x'
            ),
            r.Route(
                '/<int:page>',
                host='files.<domain>',
                endpoint='x'
            )
        ], host_matching=True)

        a = m.bind('www.example.com')
        self.assert_equal(
            a.match('/'),
            ('index', {'domain': 'example.com'})
        )
        self.assert_equal(
            a.match('/foo/'),
            ('x', {'domain': 'example.com', 'page': 1})
        )
        try:
            a.match('/foo')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://www.example.com/foo/')
        else:
            assert False, 'expected redirect'

        a = m.bind('files.example.com')
        self.assert_equal(
            a.match('/'),
            ('files', {'domain': 'example.com'})
        )
        self.assert_equal(
            a.match('/2'),
            ('x', {'domain': 'example.com', 'page': 2})
        )
        try:
            a.match('/1')
        except r.RequestRedirect as e:
            self.assert_equal(e.new_url, 'http://www.example.com/foo/')
        else:
            assert False, 'expected redirect'

    def test_server_name_casing(self):
        m = r.URLMap([
            r.Route('/', endpoint='index', subdomain='foo')
        ])

        env = create_environ()
        env['SERVER_NAME'] = env['HTTP_HOST'] = 'FOO.EXAMPLE.COM'
        a = m.bind_to_environ(env, server_name='example.com')
        self.assert_equal(a.match('/'), ('index', {}))

        env = create_environ()
        env['SERVER_NAME'] = '127.0.0.1'
        env['SERVER_PORT'] = '5000'
        del env['HTTP_HOST']
        a = m.bind_to_environ(env, server_name='example.com')
        try:
            a.match()
        except r.NotFound:
            pass
        else:
            assert False, 'Expected not found exception'

    def test_redirect_request_exception_code(self):
        exc = r.RequestRedirect('http://www.google.com/')
        exc.code = 307
        env = create_environ()
        self.assert_strict_equal(exc.get_response(env).status_code, exc.code)

    def test_redirect_path_quoting(self):
        url_map = r.URLMap([
            r.Route('/<category>', defaults={'page': 1}, endpoint='category'),
            r.Route('/<category>/page/<int:page>', endpoint='category')
        ])

        adapter = url_map.bind('example.com')
        try:
            adapter.match('/foo bar/page/1')
        except r.RequestRedirect as e:
            response = e.get_response({})
            self.assert_strict_equal(response.headers['location'],
                                     u'http://example.com/foo%20bar')
        else:
            self.fail('Expected redirect')

    def test_unicode_routes(self):
        m = r.URLMap([
            r.Route(u'/войти/', endpoint='enter'),
            r.Route(u'/foo+bar/', endpoint='foobar')
        ])
        a = m.bind(u'☃.example.com')
        try:
            a.match(u'/войти')
        except r.RequestRedirect as e:
            self.assert_strict_equal(
                e.new_url,
                'http://xn--n3h.example.com/%D0%B2%D0%BE%D0%B9%D1%82%D0%B8/'
            )
        endpoint, values = a.match(u'/войти/')
        self.assert_strict_equal(endpoint, 'enter')
        self.assert_strict_equal(values, {})

        try:
            a.match(u'/foo+bar')
        except r.RequestRedirect as e:
            self.assert_strict_equal(
                e.new_url,
                'http://xn--n3h.example.com/foo+bar/'
            )
        endpoint, values = a.match(u'/foo+bar/')
        self.assert_strict_equal(endpoint, 'foobar')
        self.assert_strict_equal(values, {})

        url = a.build('enter', {}, force_external=True)
        self.assert_strict_equal(
            url,
            'http://xn--n3h.example.com/%D0%B2%D0%BE%D0%B9%D1%82%D0%B8/'
        )

        url = a.build('foobar', {}, force_external=True)
        self.assert_strict_equal(url, 'http://xn--n3h.example.com/foo+bar/')

    def test_map_repr(self):
        m = r.URLMap([
            r.Route(u'/wat', endpoint='enter'),
            r.Route(u'/woop', endpoint='foobar')
        ])
        rv = repr(m)
        self.assert_strict_equal(
            rv,
            "URLMap([<Route '/woop' -> foobar>, <Route '/wat' -> enter>])"
        )


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RoutingTestCase))
    return suite



# -*- coding: utf-8 -*-
"""
    verktyg.exception_dispatch
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
from werkzeug.exceptions import NotAcceptable

from verktyg.accept import Representation, select_representation


class ExceptionHandlerFactory(object):
    def get_exception_handlers(self):
        raise NotImplementedError()


class ExceptionHandler(ExceptionHandlerFactory, Representation):
    """An action associated with an exception class and providing a particular
    representation.

    `exception_class`
        Binding can be used to render instances of this class and all
        subclasses

    `action`
        The action to perform if the handler is matched.
        Function accepting `(application, request, exception)` and returning
        a werkzeug response object.
    """
    def __init__(self, exception_class, action, **kwargs):

        self.exception_class = exception_class
        self.action = action

        super(ExceptionHandler, self).__init__(**kwargs)

    def get_exception_handlers(self):
        yield self

    def __repr__(self):
        return '<%s %s %s>' % (
            self.__class__.__name__,
            repr(self.exception_class),
            repr(self.content_type),
        )


class ExceptionDispatcher(ExceptionHandlerFactory):
    def __init__(self, handlers=[]):
        self._handlers = {}

        self.add_exception_handlers(*handlers)

    def add_exception_handlers(self, *factories):
        """Bind a handlers from a handler factory to render exceptions of a
        particular class or representation.
        Dispatchers can be nested

        :param factories:
            a number of instances of `ExceptionHandlerFactory` or other objects
            providing a `get_exception_handlers` method which returns an
            iterator that yields exception handlers.  Both `ExceptionHandler`
            and `ExceptionDispatcher` implement this interface so both can be
            nested using a call to `add_exception_handlers`
        """
        for factory in factories:
            for handler in factory.get_exception_handlers():
                if handler.exception_class not in self._handlers:
                    self._handlers[handler.exception_class] = []
                self._handlers[handler.exception_class].append(handler)

    def get_exception_handlers(self):
        return iter(self._handlers)

    def lookup(self, exception_class,
               accept='*/*', accept_language=None, accept_charset=None):
        """ Given an exception class and the contents of a requests accept
        headers, returns the corresponding exception handler.

        :param exception_class:
            Matched to exception handlers using `isinstance`.

        :param accept:
            See `verktyg.accept.select_representation` for details

        :param accept_language:
            See `verktyg.accept.select_representation` for details

        :param accept_charset:
            See `verktyg.accept.select_representation` for details

        :return:
            A callable object accepting am application object, a request, and
            an exception and returning a werkzeug response
        """
        # Use the method resolution order of the exception to rank handlers
        for cls in exception_class.mro():
            if cls not in self._handlers:
                continue

            try:
                handler = select_representation(
                    self._handlers[cls],
                    accept=accept,
                    accept_language=accept_language,
                    accept_charset=accept_charset
                )

                return handler.action

            except NotAcceptable:
                continue

        return None



"""
    verktyg.testsuite.application
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests `Application` utility class.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from werkzeug.test import Client
from werkzeug.testsuite import WerkzeugTestCase

from werkzeug import Response, BaseResponse
from werkzeug.exceptions import HTTPException, NotFound, ImATeapot

from verktyg.views import expose
from verktyg.routing import Route
from verktyg.application import Application


class ApplicationTestCase(WerkzeugTestCase):
    def test_basic(self):
        app = Application()

        app.add_routes(Route('/', endpoint='index'))

        @expose(app.dispatcher, 'index')
        def index(app, request):
            return Response('Hello World')

        client = Client(app, BaseResponse)

        resp = client.get('/')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.get_data(), b'Hello World')

    def test_exception_handlers(self):
        app = Application()

        @app.exception_handler(BaseException)
        def default_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('default handler', status=500)

        @app.exception_handler(HTTPException)
        def werkzeug_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('werkzeug handler', exc_value.code)

        @app.expose(route='/raise_execption')
        def raise_exception(app, req):
            raise Exception()

        @app.expose(route='/raise_key_error')
        def raise_key_error(app, req):
            raise KeyError()

        @app.expose(route='/raise_teapot')
        def raise_teapot(app, req):
            raise ImATeapot()

        client = Client(app, BaseResponse)

        resp = client.get('/raise_execption')
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.get_data(), b'default handler')

        resp = client.get('/raise_key_error')
        self.assertEqual(resp.status_code, 500)
        self.assertEqual(resp.get_data(), b'default handler')

        resp = client.get('/raise_teapot')
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'werkzeug handler')

    def test_middleware(self):
        app = Application()

        @app.expose(route='/')
        def index(app, req):
            return Response()

        results = dict(
            got_request=False,
            got_response=False,
        )

        def middleware(app):
            def handler(env, start_response):
                results['got_request'] = True

                def handle_start_response(*args, **kwargs):
                    results['got_response'] = True
                    return start_response(*args, **kwargs)

                app(env, handle_start_response)
            return handler

        app.add_middleware(middleware)

        client = Client(app, BaseResponse)

        client.get('/')

        self.assertTrue(results['got_request'])
        self.assertTrue(results['got_response'])

    def test_exception_content_type(self):
        app = Application()

        @app.exception_handler(HTTPException)
        def default_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response('default handler', status=exc_value.code)

        @app.exception_handler(HTTPException, content_type='application/json')
        def default_json_handler(app, req, exc_type, exc_value, exc_traceback):
            return Response(
                '{"type": "json"}',
                status=exc_value.code,
                content_type='application/json'
            )

        @app.exception_handler(NotFound, content_type='text/html')
        def html_not_found_handler(
                app, req, exc_type, exc_value, exc_traceback):
            return Response('pretty NotFound', status=exc_value.code)

        @app.expose(route='/raise_418')
        def raise_418(app, req):
            raise ImATeapot()

        @app.expose(route='/raise_404')
        def raise_404(app, req):
            raise NotFound()

        client = Client(app, BaseResponse)

        resp = client.get('/raise_418')
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'default handler')

        resp = client.get(
            'raise_418', headers=[('Accept', 'application/json')]
        )
        self.assertEqual(resp.status_code, 418)
        self.assertEqual(resp.get_data(), b'{"type": "json"}')
        self.assertEqual(resp.headers['Content-Type'], 'application/json')

        # 404 error has a pretty html representation but uses default renderer
        # for json
        resp = client.get('/raise_404', headers=[('Accept', 'text/html')])
        self.assertEqual(resp.status_code, 404)
        self.assertEqual(resp.get_data(), b'pretty NotFound')

        resp = client.get(
            'raise_404', headers=[('Accept', 'application/json')]
        )
        self.assertEqual(resp.status_code, 404)
        self.assertEqual(resp.get_data(), b'{"type": "json"}')
        self.assertEqual(resp.headers['Content-Type'], 'application/json')

    def test_properties(self):
        app = Application()

        # test getters
        app.add_property('foo', lambda self: 'bar')

        self.assertEqual(app.foo, 'bar')

        try:
            app.foo = 'baz'
        except AttributeError:
            pass
        else:
            self.fail("Should raise AttributeError")

        # test setters
        def geter(self):
            if hasattr(self, '_fizz'):
                return self._fizz * 2
            else:
                return 'default'

        def seter(self, value):
            self._fizz = value

        app.add_property('fizz', geter, seter)

        self.assertEqual(app.fizz, 'default')

        app.fizz = 2

        self.assertEqual(app.fizz, 4)

    def test_methods(self):
        app = Application()

        app.add_method('get_self', lambda self: self)
        self.assertEqual(app.get_self(), app)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ApplicationTestCase))
    return suite



"""
    verktyg.testsuite.accept
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for choosing representations based on content type, etc.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from werkzeug.testsuite import WerkzeugTestCase

from verktyg.accept import Representation


class RepresentationTestCase(WerkzeugTestCase):
    def test_content_type_dispatch(self):
        default_binding = Representation()
        html_binding = Representation(content_type='text/html')

        self.assert_equal(
            (5, 0.001),
            default_binding.quality(accept='text/html'))

        self.assert_equal(
            (110, 1.0),
            html_binding.quality(accept='text/html'))

        self.assert_equal(
            (110, 0.9),
            html_binding.quality(accept='text/html;q=0.9'))

        self.assert_equal(
            (0, 0.01),
            html_binding.quality(accept='application/json, */*; q=0.01'))


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RepresentationTestCase))
    return suite



"""
    verktyg.testsuite.views
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for view wrappers.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
from werkzeug.testsuite import WerkzeugTestCase

from werkzeug.wrappers import Response
from werkzeug.exceptions import MethodNotAllowed

import verktyg as d


class ViewsTestCase(WerkzeugTestCase):
    def test_decorators(self):
        dispatcher = d.Dispatcher()

        @d.expose(dispatcher, 'foo')
        def foo(env, req):
            pass

    def test_class_view(self):
        dispatcher = d.Dispatcher()

        class Foo(d.ClassView):
            name = 'foo'

            def GET(self, env, req):
                return 'get'

            def POST(self, env, req):
                return 'post'

        dispatcher.add(Foo())

        self.assert_equal('get',
                          dispatcher.lookup('foo', method='GET')(None, None))
        self.assert_equal('post',
                          dispatcher.lookup('foo', method='POST')(None, None))
        self.assert_raises(MethodNotAllowed,
                           dispatcher.lookup, 'foo', method='PUT')

    def test_template_view(self):
        dispatcher = d.Dispatcher()

        class HelloEnv(object):
            def get_renderer(self, name):
                if name == 'hello':
                    return lambda res: Response('hello %s' % res)
        env = HelloEnv()

        @d.expose(dispatcher, 'say-hello', template='hello')
        def say_hello(env, req):
            return 'world'

        @d.expose(dispatcher, 'returns-response', template='hello')
        def returns_response(env, req):
            return Response('too slow')

        self.assert_equal(
            b'hello world',
            dispatcher.lookup('say-hello')(env, None).get_data())
        self.assert_equal(
            b'too slow',
            dispatcher.lookup('returns-response')(env, None).get_data())
